""" Set of available and useful LLMs (mostly posted on AWS Bedrock)
"""


class LLM_Service:
    def __str__(self):
        return self.llm_description

    def __repr__(self):
        return self.llm_description

    def __call__(
        self,
        msg,
        b64images=None,
        system_prompt="You are a helpful assistant. Do not use emojis in the answers.",
        chat_history=[],
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=3,
        cur_fail_sleep=60,
    ):
        """Calls the LLM in streaming mode
        Arguments:
        system_prompt: prompt that should persist across questions, using specialist attention
        msg: next user message
        chat_history: list of lists. Each inner element should contain [<user msg>, <assistant msg>]
        """
        assert isinstance(
            extra_stop_sequences, list
        ), "extra_stop_sequences should be a list of strings"
        call_list = self._prepare_call_list_from_history(
            system_prompt, msg, b64images, chat_history
        )
        prompt = self._prepare_prompt_from_list(call_list)
        self.last_prompt = str(prompt) + postpend
        if tools is None:
            return self.invoke_streaming(
                prompt,
                postpend=postpend,
                extra_stop_sequences=extra_stop_sequences,
                max_retries=max_retries,
                cur_fail_sleep=cur_fail_sleep,
            )
        else:
            return self.invoke_streaming(
                prompt,
                postpend=postpend,
                extra_stop_sequences=extra_stop_sequences,
                tools=tools,
                tool_invoker_fn=tool_invoker_fn,
                max_retries=max_retries,
                cur_fail_sleep=cur_fail_sleep,
            )

    def _prepare_call_list_from_history(
        self,
        system_prompt,
        msg,
        b64images,
        chat_history,
        context_reset_string="[|[PAST_FORGOTTEN]|]",
    ):
        """Prepares the prompt for the next interaction with the LLM.
        This image preparation is suited for Anthropic's Claude
        """
        history_list = [
            {"role": "system", "content": system_prompt},
        ]
        for x in chat_history:
            # allows the LLM to determine that the past is no longer relevant, to free up context
            if context_reset_string in str(x):
                history_list = [
                    history_list[0],
                    {
                        "role": "user",
                        "content": str(x),
                    },
                ]
            elif isinstance(x, dict):
                history_list.append(x)
            else:
                history_list.append({"role": "user", "content": x[0]})
                history_list.append({"role": "assistant", "content": str(x[1])})

        if b64images is None:
            history_list.append({"role": "user", "content": msg})
        else:
            if not isinstance(b64images, list):
                b64images = [b64images]
            cur_content = []
            for b64img in b64images:
                if b64img is not None:
                    cur_content.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": "image/jpeg",
                                "data": b64img,
                            },
                        },
                    )
            cur_content.append({"type": "text", "text": msg})
            history_list.append({"role": "user", "content": cur_content})
        return history_list



from unittest.mock import call, patch, mock_open

from gat_llm.tools.read_local_file import ToolReadLocalFile


def test_unexpected_arg(unexpected_param_msg):
    trlf = ToolReadLocalFile()
    ans = trlf("media/file.txt", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


@patch("builtins.open", new_callable=mock_open, read_data="file_data")
@patch("os.path.isfile", return_value=True)
def test_file_read(mock_isfile, mock_open_file):
    trlf = ToolReadLocalFile()
    ans = trlf("media/file.txt")

    mock_isfile.assert_called_with("media/file.txt")
    assert "file_data" in ans


@patch("builtins.open", new_callable=mock_open, read_data="file_data")
@patch("os.path.isfile", return_value=True)
def test_files_read(mock_isfile, mock_open_file):
    trlf = ToolReadLocalFile()
    ans = trlf("media/file.txt \n media/file2.txt")

    mock_isfile.assert_has_calls([call("media/file.txt"), call("media/file2.txt")])
    assert "file_data" in ans


@patch("builtins.open", new_callable=mock_open, read_data="file_data")
@patch("os.path.isfile", return_value=False)
def test_file_not_read(mock_isfile, mock_open_file):
    trlf = ToolReadLocalFile()
    ans = trlf("media/file.txt")

    mock_isfile.assert_called_with("media/file.txt")
    mock_open_file.assert_not_called()
    assert "Error: Did not find file" in ans



import os

from openai import OpenAI
from .openai import LLM_GPT_OpenAI


class LLM_Deepseek(LLM_GPT_OpenAI):
    def __init__(self, model_size):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        if model_size == "Deepseek Chat":
            self.model_id = "deepseek-chat"
            self.llm_description = (
                "Deepseek V3 (medium-sized LLM) - directly from DeepSeek"
            )
            self.price_per_M_input_tokens = 0.27
            self.price_per_M_output_tokens = 1.1

        try:
            self.openai_client = OpenAI(
                api_key=os.environ.get("DEEPSEEK_API_KEY"),
                base_url="https://api.deepseek.com",
            )
        except Exception:
            self.openai_client = None

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []



class ToolSummarizePast:
    def __init__(self):
        self.name = "summarize_past_on_context_switch"

        self.tool_summary = f"""<tool_summary>
<tool_name>{self.name}</tool_name>
<summary>Summarizes all previous conversations when the user switches context, to avoid keeping in memory a very long conversation that is no longer relevant.</summary>
</tool_summary>"""

        self.tool_description = {
            "name": self.name,
            "description": f"""Summarizes all previous conversations when the user switches context, to avoid keeping in memory a very long conversation that is no longer relevant.

Before calling any other tool, evaluate if {self.name} should be called.
Before every answer, use the <scratchpad></scratchpad> to evaluate if the next answer is unrelated to the previous topics.

CRITICAL: This tool is essential for maintaining conversation efficiency and relevance. Use it proactively to manage memory and context.

WHEN TO USE:
1. After every 3-5 exchanges, evaluate if a summary is needed.
2. When the user introduces a new topic unrelated to the main themes of the last 3-5 exchanges.
3. If the conversation has drifted significantly from the original topic.
4. When the user explicitly indicates a topic change (e.g., "Let's talk about something else", "Changing the subject").
5. If a certain keyword analysis threshold is met (see below).

PROCESS:
1. Maintain a brief list of main topics discussed, updating with each exchange.
2. Before each response, use the <scratchpad></scratchpad> to:
   a. Perform a keyword analysis between the current query and the previous 2-3 exchanges.
   b. Evaluate if the next answer will be unrelated to the previous topics.
   c. Consider what information might be relevant for future exchanges.

3. If a topic change is detected or 5 exchanges have passed, explicitly explain in the scratchpad whether summarize_past_on_context_switch will be used and why.

4. When calling the tool, ensure to:
   a. Provide a concise summary of important topics discussed.
   b. Include crucial information from both user and assistant.
   c. Specify information that may still be relevant moving forward.
   d. State the next question to be answered.

KEYWORD ANALYSIS:
Perform a simple comparison of key terms between the current query and recent exchanges. If the overlap is less than 30%, consider it a potential topic change.

EXAMPLES:
- Topic change: From discussing solar system to cooking recipes.
- Not a topic change: From discussing planets to discussing moons (related astronomical objects).

THRESHOLD:
Consider a topic "unrelated" if it doesn't share any major concepts, field of study, or direct implications with the previous topic.

IMPORTANCE:
Proper use of this tool is as crucial as using tools for information retrieval or computation. It directly impacts the quality and efficiency of the conversation.

Failing to use this tool when appropriate will result in less efficient and less relevant conversations, and may be flagged for improvement.

Raises ValueError: if one of the parameters is invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "past_conversation_summary": {
                        "type": "string",
                        "description": "A summary of important topics that were discussed. Always include important information from user and assistant (you).",
                    },
                    "information_still_relevant": {
                        "type": "string",
                        "description": "Information that may still be relevant moving forward with the conversation, like user preferences, name, etc.",
                    },
                    "next_question_to_answer": {
                        "type": "string",
                        "description": "What was the last question from the user, which still needs to be answered.",
                    },
                },
                "required": [
                    "past_conversation_summary",
                    "information_still_relevant",
                    "next_question_to_answer",
                ],
            },
        }
        self.tool_summary = self.tool_description

    def __call__(
        self,
        past_conversation_summary,
        information_still_relevant,
        next_question_to_answer,
        **kwargs,
    ):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        ans = f"""Summary of all past conversation generated by tool {self.name}:

<past_conversation_summary>{past_conversation_summary}</past_conversation_summary>
<information_still_relevant>{information_still_relevant}</information_still_relevant>
<next_question_to_answer>{next_question_to_answer}</next_question_to_answer>

[|[PAST_FORGOTTEN]|]
"""
        # this tool doesn't really need to do anything. The arguments are what is important
        return ans



import os

from openai import OpenAI
from .openai import LLM_GPT_OpenAI


class LLM_Maritalk(LLM_GPT_OpenAI):
    def __init__(self, model_size):
        """Constructor
        Arguments:
            model_size - Maritaca model to use to make LLM calls
        """
        if model_size == "Sabia3 Maritaca":
            self.model_id = "sabia-3.1"
            self.llm_description = (
                "Sabia-3.1 (medium-sized LLM) - directly from Maritaca"
            )
            self.price_per_M_input_tokens = 0.95
            self.price_per_M_output_tokens = 1.9

        try:
            self.openai_client = OpenAI(
                api_key=os.environ.get("MARITACA_API_KEY"),
                base_url="https://chat.maritaca.ai/api",
            )
        except Exception:
            self.openai_client = None

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []



import os


class ToolWriteLocalFile:
    def __init__(self):
        self.name = "write_local_files"

        self.tool_description = {
            "name": self.name,
            "description": """Write a text file in the local file system.
Only write text files, like files with extensions .txt, .py, .md and others that usually contain only text.
Do not attempt to write files that are usually in binary format. If the path is not specified, write the file to the media/ folder.

Raises ValueError: if the file could not be written.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "path_to_file": {
                        "type": "string",
                        "description": "Local path to the file which will be written.",
                    },
                    "text_content": {
                        "type": "string",
                        "description": "Content to write in the file (text).",
                    },
                },
                "required": ["path_to_file", "text_content"],
            },
        }

    def __call__(self, path_to_file, text_content, **kwargs):
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        final_ans = ["<outcome>"]
        if not os.path.isfile(path_to_file):
            with open(path_to_file, "w", encoding="utf-8") as f:
                f.write(text_content)
                final_ans.append("File written successfully")
                final_ans.append(f"<path_to_file>{path_to_file}</path_to_file>")
        else:
            final_ans.append(
                f"Error: File already exists: {path_to_file}. This tool is not allowed to overwrite files."
            )
        final_ans.append("</outcome>")
        return "\n".join(final_ans)



import os
import re
import json
import time
import uuid
import base64
from io import BytesIO

import numpy as np
from PIL import Image


def _adjust_msg_for_gradio_ui(x, show_scratchpad=False, show_calls=False):
    """Adjusts a string to be displayed in the gradio UI

    <sup><sub>This is subbed-sup text</sub></sup>
    """
    # <!--This is a comment. Comments are not displayed in the browser-->
    # return x.replace('<answer>', '\n\nAnswer:').replace('</answer>', '')
    # x = x.replace('<', '[').replace('>', ']')

    if show_scratchpad:
        x = x.replace("<scratchpad>", "```\n[scratchpad]")
        x = x.replace("</scratchpad>", "[/scratchpad]\n```\n")
        x = x.replace("<think>", "```\n[think]")
        x = x.replace("</think>", "[/think]\n```\n")
    else:
        # System.Text.RegularExpressions.Regex.Replace(test_str,"<a>[\S\s]*?</a>\s*", "")
        x = re.sub(r"<scratchpad>[\S\s]*?</scratchpad>\s*", "", x)
        x = re.sub(r"<think>[\S\s]*?</think>\s*", "", x)
        # x = x.replace('<scratchpad>', '<!--<scratchpad>')
        # x = x.replace('</scratchpad>', '</scratchpad>-->')

    if not show_calls:
        x = re.sub(r"<function_calls>[\S\s]*?</function_calls>\s*", "", x)

    x = re.sub(r"<function_results>[\S\s]*?</function_results>\s*", "", x)
    # x = x.replace('<function_results>', '<!--<function_results>')
    # x = x.replace('</function_results>', '<function_results>-->')

    x = x.replace("<answer>", "<answer><b>").replace("</answer>", "</answer></b>")
    return x


class LLMInterface:
    def __init__(
        self,
        system_prompt,
        llm,
        llm_tools,
        rpg,
        output_mode="chat_bot",
        chat_log_folder="chat_logs",
        show_extra_info=True,
    ):
        """Constructor

        Arguments
        system_prompt: desired system prompt
        llm: LLM to use. Method invoke_streaming will be used
        llm_tools: implementation of resolutor to LLM tool calls. Has to implement invoke_from_cmd
        rpg: Instance of RAGPromptGenerator. Needs to have post_anti_hallucination property
        output_mode: chat_bot uses Gradio customized chatbot that has to
            receive the whole history. Otherwise uses gradio chatinterface
        chat_log_folder: folder to save chat to. If None, does not save chat
        show_extra_info: show extra info like tools used and scratchpad in the UI?
        """
        self.system_prompt = system_prompt
        self.llm = llm
        self.lt = llm_tools
        self.rpg = rpg
        self.chat_log_folder = chat_log_folder
        self.show_extra_info = show_extra_info

        # keep a hash of histories so we can send to the UI
        # something different than what has been generated
        self.history_log = {}

        valid_output_modes = ["chat_interface", "chat_bot"]
        assert (
            output_mode in valid_output_modes
        ), f"Output mode must be one of {valid_output_modes}"
        self.output_mode = output_mode
        # if requested by the LLM, get rid of the past history
        self.erase_past = False
        # keep some execution logs
        self.log = []

        # handle native tool use
        if self.rpg is not None and self.rpg.use_native_tools:
            self.native_tools = [x.tool_description for x in self.lt.tools]
            self.tool_invoker_fn = self.lt.invoke_tool
            self.extra_stop_sequences = []
        else:
            self.native_tools = None
            self.tool_invoker_fn = None
            self.extra_stop_sequences = ["</function_calls>"]

    def _format_msg(
        self, x, message, chat_history, show_ans_only=False, extra_info=None
    ):
        # the "<path_to_" substring from native tools has to be appended to the final answer for file display
        if self.output_mode == "chat_interface":
            return _adjust_msg_for_gradio_ui(x)
        elif self.output_mode == "chat_bot":
            cur_history = chat_history.copy()
            cur_ans = _adjust_msg_for_gradio_ui(x)

            ans_only = cur_ans.split("<answer>")
            if len(ans_only) > 1 and show_ans_only:
                cur_ans = ans_only[-1]
            elif show_ans_only:
                cur_ans = ""

            info_msg = None
            if extra_info is None or not self.show_extra_info:
                cur_history += [
                    {"role": "user", "content": message},
                    {"role": "assistant", "content": cur_ans},
                ]
            else:
                info_msg = {
                    "role": "assistant",
                    "content": extra_info.get("content", ""),
                    "metadata": extra_info["metadata"],
                }
                cur_history += [
                    {"role": "user", "content": message},
                    info_msg,
                    {"role": "assistant", "content": cur_ans},
                ]

            # find out if we should be showing an image
            image_candidates = x.split("<path_to_image>")
            shown_images = {}
            for k in range(1, len(image_candidates)):
                image_candidate = image_candidates[k].split("</path_to_image>")[0]

                if os.path.isfile(image_candidate) and not shown_images.get(
                    image_candidate, False
                ):
                    cur_history.append(
                        {
                            "role": "assistant",
                            "content": {"path": image_candidate, "alt_text": "media"},
                        }
                    )
                    shown_images[image_candidate] = True

            # find out if we should be showing an audio
            audio_candidates = x.split("<path_to_audio>")
            shown_audios = {}
            for k in range(1, len(audio_candidates)):
                audio_candidate = audio_candidates[k].split("</path_to_audio>")[0]

                if os.path.isfile(audio_candidate) and not shown_audios.get(
                    audio_candidate, False
                ):
                    cur_history.append(
                        {
                            "role": "assistant",
                            "content": {"path": audio_candidate, "alt_text": "media"},
                        }
                    )
                    shown_audios[audio_candidate] = True

            # find out if we should add a downloadable file
            file_candidates = x.split("<path_to_file>")
            shown_files = {}
            for k in range(1, len(file_candidates)):
                file_candidate = file_candidates[k].split("</path_to_file>")[0]
                if os.path.isfile(file_candidate) and not shown_files.get(
                    file_candidate, False
                ):
                    cur_history.append(
                        {
                            "role": "assistant",
                            "content": {"path": file_candidate, "alt_text": "media"},
                        }
                    )
                    shown_files[file_candidate] = True

            # figure out what should go into the scratchpad
            scratchpad_info = ""
            if "<scratchpad>" in x:
                scratchpad_info = x.split("<scratchpad>")
                if len(scratchpad_info) > 1:
                    scratchpad_info = scratchpad_info[-1].split("</scratchpad>")[0]
            elif "<think>" in x:
                scratchpad_info = x.split("<think>")
                if len(scratchpad_info) > 1:
                    scratchpad_info = scratchpad_info[-1].split("</think>")[0]

            if info_msg is not None:
                info_msg["content"] = (
                    extra_info.get("content", "") + "\n" + scratchpad_info
                )

            # also put function calls in there
            func_call_info = x.split("<function_calls>")
            if len(func_call_info) > 1:
                scratchpad_info = (
                    scratchpad_info
                    + "\n\nFunction call:\n\n"
                    + func_call_info[-1].split("</function_calls>")[0]
                )

            # make sure to send ChatBot history last
            return "", scratchpad_info, None, cur_history

    def _rem_none(self, history):
        """Returns a copy of history but with None messages from users removed"""
        return str([x for x in history if x[0] is not None])

    def chat_with_function_caller(self, msg, images, ui_history=[], username=""):
        """Performs conversation with the LLM agent

        Arguments:
            msg: next user message
            image: user input image
            ui_history: history in the user interface. The first is used to recover the state in memory
            username: user name of the user logged in the UI
        """
        image_strings = None
        if images is not None:
            image_strings = []
            if not isinstance(images, list):
                images = [images]
            for image in images:
                if image is None:
                    image_strings.append(None)
                else:
                    npimg = np.array(image, dtype=np.uint8)
                    pil_img = Image.fromarray(npimg)
                    buff = BytesIO()
                    pil_img.save(buff, format="JPEG")
                    image_string = base64.b64encode(buff.getvalue()).decode("utf-8")
                    image_strings.append(image_string)

        t0 = time.time()

        if len(ui_history) > 0:
            chat_id = ui_history[0]["content"]
            history = self.history_log[chat_id]
            # with open('ui_debug.txt', 'w') as f:
            #    f.write(str([msg, history]))
        else:
            chat_id = str(uuid.uuid4())
            ui_history = [{"role": "assistant", "content": chat_id}]
            history = []

        ans2 = self.llm(
            msg,
            b64images=image_strings,
            system_prompt=self.system_prompt,
            chat_history=history,
            postpend=self.rpg.post_anti_hallucination
            if (self.rpg is not None and not self.rpg.use_native_tools)
            else "",
            extra_stop_sequences=self.extra_stop_sequences,
            tools=self.native_tools,
            tool_invoker_fn=self.lt.invoke_tool if self.lt is not None else None,
        )

        extra_info = {
            "metadata": {"title": "🧠", "status": "pending"},
        }
        x = ""
        for x in ans2:
            if self.lt is not None and hasattr(self.llm, "cur_tool_spec"):
                extra_info = {
                    "metadata": {"title": "🛠️", "status": "pending"},
                    "content": ", ".join([x["tool_name"] for x in self.lt.invoke_log]),
                }
            yield self._format_msg(x, msg, ui_history, extra_info=extra_info)
        # initial_ans = self._format_msg(x, msg, ui_history)
        # yield initial_ans

        cur_answer = x

        cur_answer_split = cur_answer.split("<function_calls>")
        while (
            len(cur_answer_split) > 1
            and self.lt is not None
            and not self.rpg.use_native_tools
        ):
            # this loop means that manual tool usage is needed
            cur_func_log = {}

            xml_to_parse = cur_answer_split[-1].split("</function_calls>")[0]
            post_prompt = self.lt.invoke_from_cmd(xml_to_parse, username=username)

            cur_func_log["Parse and exec query"] = {
                "exec_time": time.time() - t0,
            }
            t0 = time.time()

            cur_postpend = cur_answer + post_prompt
            # yield self._format_msg(x, msg, ui_history)

            # note: parameters tools and tool_invoker_fn are not used
            # because this call fulfills manual tool use requests
            # ie. if there are native tools, this loop should never happen
            ans2 = self.llm(
                msg,
                system_prompt=self.system_prompt,
                chat_history=history,
                postpend=cur_postpend if not self.rpg.use_native_tools else "",
                extra_stop_sequences=self.extra_stop_sequences,
            )

            for x in ans2:
                # pass
                yield self._format_msg(x, msg, ui_history)
            # yield self._format_msg(x, msg, ui_history)

            log_dict = self.llm.word_counts[-1].copy()
            cur_func_log["Analyze query with LLM"] = {
                "exec_time": time.time() - t0,
                "word_count": log_dict,
            }
            t0 = time.time()
            cur_answer = x
            last_update = cur_answer.replace(cur_postpend, "")
            cur_answer_split = last_update.split("<function_calls>")

        history_to_append = []
        tool_results = []
        if hasattr(self.llm, "tool_use_added_msgs"):
            history_to_append.append({"role": "user", "content": msg})
            tool_results.append("\n")
            for x in self.llm.tool_use_added_msgs:
                history_to_append.append(x)

                # enable media display in the Gradio UI - amazon Nova
                if x["role"] == "user" and x["content"][0].get("toolResult"):
                    cur_tool_result = x["content"][0]["toolResult"]["content"][0][
                        "text"
                    ]
                    tool_results.append(
                        cur_tool_result if "<path_to_" in cur_tool_result else ""
                    )
                # enable media display in the Gradio UI - anthropic
                if x["role"] == "user" and x["content"][0].get("content"):
                    cur_tool_result = x["content"][0]["content"]
                    tool_results.append(
                        cur_tool_result if "<path_to_" in cur_tool_result else ""
                    )
                # enable media display in the Gradio UI - openai
                if x["role"] == "tool":
                    cur_tool_result = x["content"]
                    tool_results.append(
                        cur_tool_result if "<path_to_" in cur_tool_result else ""
                    )
            history_to_append.append({"role": "assistant", "content": cur_answer})
        else:
            history_to_append.append([msg, cur_answer])

        tool_results = "\n".join(tool_results)
        self.history_log[chat_id] = history + history_to_append

        try:
            chat_log_dir = self.chat_log_folder
            if username is not None and username.strip() != "":
                cur_user_to_log = username
            else:
                cur_user_to_log = "unknown"
            if chat_log_dir is not None:
                os.makedirs(chat_log_dir, exist_ok=True)
                with open(
                    f"{chat_log_dir}/{cur_user_to_log}-{chat_id}.json",
                    "w",
                    encoding="utf-8",
                ) as f:
                    f.write(json.dumps(self.history_log[chat_id]))
        except Exception as ex:
            print(
                f"Could not log chat to folder `{self.chat_log_folder}`. Reason: {str(ex)}"
            )

        extra_info["metadata"]["status"] = "done"
        final_response_ui = self._format_msg(
            cur_answer + tool_results, msg, ui_history, extra_info=extra_info
        )
        if self.lt is not None:
            self.lt.invoke_log = []
        yield final_response_ui
        print("Final response sent.")



import os
import json
import base64
from io import BytesIO

import numpy as np
from PIL import Image
from openai import OpenAI

rng = np.random.default_rng()


def remove_semi_transparent_pixels(image_path: str) -> bytes:
    """
    Opens an RGBA image, sets all pixels with alpha != 255 to fully transparent black (0,0,0,0)
    and returns the resulting PNG image as bytes.

    :param image_path: Path to the input image.
    :return: PNG-encoded image bytes.
    """
    # Load the image as RGBA
    img = Image.open(image_path).convert("RGBA")

    # Convert to NumPy array
    arr = np.array(img)

    # Mask for alpha not equal to 255
    alpha_mask = arr[:, :, 3] != 255
    not_alpha_mask = arr[:, :, 3] == 255

    # Set affected pixels to fully transparent black
    arr[alpha_mask] = [0, 0, 0, 0]
    arr[not_alpha_mask] = [0, 0, 0, 255]

    # Convert back to PIL Image
    result_img = Image.fromarray(arr, mode="RGBA")

    # Save to bytes
    buf = BytesIO()
    result_img.save(buf, format="PNG")
    mask_bytes = buf.getvalue()
    # Save the resulting file
    with open(image_path, "wb") as f:
        f.write(mask_bytes)


class ToolImageEdit:
    def _gen_img_openai(self, image_paths, prompt, mask_file, target_file):
        if self.openai_client is None:
            self.openai_client = OpenAI()

        args = {
            "model": "gpt-image-1",
            "image": [open(x, "rb") for x in image_paths],
            "prompt": prompt,
            "n": 1,
        }
        if mask_file is not None:
            remove_semi_transparent_pixels(mask_file)
            args["mask"] = open(mask_file, "rb")

        response = self.openai_client.images.edit(**args)
        parsed_ans = json.loads(response.json())
        revised_prompt = parsed_ans["data"][0]["revised_prompt"]
        img_content = base64.b64decode(parsed_ans["data"][0]["b64_json"])
        with open(target_file, "wb") as f:
            f.write(img_content)
        return revised_prompt

    def __init__(self):
        self.name = "edit_image"

        self.tool_description = {
            "name": self.name,
            "description": """Creates an edited or extended image given one or more source images and a prompt. The prompt should be precise and complete. In case of doubt, ask clarification questions to the user before calling this tool. Never mention file names in the prompt because the image generation model only receives the bytes of each image,
In all cases, including when using a mask, the prompt has to describe the entire resulting image, not just the area that is masked. When in doubt, ask the user to generate or clarify the full description.

Raises ValueError: if not able to generate the image.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "image_files": {
                        "type": "string",
                        "description": """A list containing the full path of the images to be edited. Use one file name per line. Up to 16 images can be used. Note that the order matters - the Nth image should be referred to using its number and a brief description.
These files will be used as input. For example:
path/to/image1.png
path/to/image2.jpg
""",
                    },
                    "prompt": {
                        "type": "string",
                        "description": "A complete, descriptive prompt providing precise instructions about how to edit the images and what to generate as output.",
                    },
                    "mask_file": {
                        "type": "string",
                        "description": "Full path to the image that will be used as mask. The fully transparent areas (e.g. where alpha is zero) indicate where image should be edited. If there are multiple images provided, the mask will be applied on the first image.",
                    },
                },
                "required": ["image_files", "prompt"],
            },
        }
        self.openai_client = None

    def __call__(
        self,
        image_files,
        prompt,
        mask_file=None,
        engine=None,
        **kwargs,
    ):
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/edt_image_{rng_num}.png"

        if engine is None:
            engine = "openai-img"
        engine = engine.lower()

        try:
            if engine == "openai-img":
                image_files = [x.strip() for x in image_files.splitlines()]
                for f in image_files:
                    if not os.path.isfile(f):
                        return (
                            f"Error: image file not found: {f}. Please check the path."
                        )
                self._gen_img_openai(image_files, prompt, mask_file, target_file)
        except Exception as e:
            return f"Image was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        ans = ["<image>"]
        ans.append(
            f"<used_engine>{engine}</used_engine><path_to_image>{target_file}</path_to_image>"
        )
        ans.append("</image>")
        return "\n".join(ans)



from openai import OpenAI
from .openai import LLM_GPT_OpenAI


class LLM_Ollama(LLM_GPT_OpenAI):
    def __init__(self, model):
        """Constructor
        Arguments:
            model - Desired model
        """
        # there would be electricity cost for running a model locally
        # but for now we keep cost at 0
        self.price_per_M_input_tokens = 0
        self.price_per_M_output_tokens = 0
        if model == "DeepSeek R1 14b Ollama":
            self.model_id = "deepseek-r1:14b"
            self.llm_description = (
                "DeepSeek R1 14b (Tiny-size LLM) - locally from Ollama"
            )
        elif model == "Qwen 2.5vl 7b Ollama":
            self.model_id = "qwen2.5vl:7b"
            self.llm_description = "Qwen 2.5vl 7b (Tiny-size LLM) - locally from Ollama"
        elif model == "Qwen 3 8b Ollama":
            self.model_id = "qwen3:8b"
            self.llm_description = "Qwen 3 8b (Small-size LLM) - locally from Ollama"
        elif model == "Qwen 3 14b Ollama":
            self.model_id = "qwen3:14b"
            self.llm_description = "Qwen 3 14b (Small-size LLM) - locally from Ollama"
        elif model == "Llama4 16x17b Ollama":
            # this model may be too big for many systems
            self.model_id = "llama4:16x17b"
            self.llm_description = (
                "Llama 4 16x17b (Small-size LLM) - locally from Ollama"
            )

        try:
            self.openai_client = OpenAI(
                base_url="http://localhost:11434/v1/",
                api_key="ollama",  # required but ignored
            )
        except Exception:
            self.openai_client = None

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []



import pytest
from unittest.mock import patch, Mock
from gat_llm.tools.get_webpage_contents import ToolGetUrlContent


@pytest.fixture
def mock_requests_get():
    with patch("requests.get") as mock_get:
        mock_response = Mock()
        mock_response.content = (
            "<html><body><p><div>Test content</div></p></body></html>"
        )
        mock_response.text = "Test content"
        mock_response.url = "http://example.com"
        mock_response.status_code = 200
        mock_get.return_value = mock_response
        yield mock_get


def test_unexpected_arg(unexpected_param_msg):
    tguc = ToolGetUrlContent(None)
    ans = tguc("http://example.com", True, unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_get_url_content_all_visible_html(mock_requests_get):
    tguc = ToolGetUrlContent(None)
    result = tguc("http://example.com", return_all_visible_html="True")
    assert "Test content" in result
    assert "<source_url>http://example.com</source_url>" in result
    assert "<status_code>200</status_code>" in result


def test_get_url_content_text_and_urls(mock_requests_get):
    tguc = ToolGetUrlContent(None)
    result = tguc("http://example.com", return_all_visible_html="False")
    assert "Test content" in result
    assert "<source_url>http://example.com</source_url>" in result
    assert "<status_code>200</status_code>" in result
    assert "<urls>" in result


@patch("requests.get", side_effect=Exception("Connection error"))
def test_get_url_content_error(mock_requests_get):
    tguc = ToolGetUrlContent(None)
    result = tguc("http://example.com")
    assert "Could not retrieve page from URL" in result
    assert "Connection error" in result


# Add more tests for different scenarios and edge cases



# Thanks to https://www.kaggle.com/datasets/kyanyoga/sample-sales-data
# Sample CSV from that page (adjusted)
"""
class llm_db has to have

 - description of all tables
 - description of all columns of all tables
 - JOIN columns (usually IDs)

the implementation will probably involve a WHERE with the account_id

so llm_db implements

llm_db.sql_query(q)

and does something like

query = WITH tbl_name as
(
	SELECT * FROM real_tbl
	WHERE account_id = (replace with account_id)
)

{q}

q will SELECT FROM tbl_name

in the LLM tool itself, possible arguments:

table: table to select from
columns_without_aggregation: list of columns, separated by commas, desired in the output whose contents will not be aggregated
aggregated_columns: list of columns, separated by commas, that will be aggregated
aggregation_functions: list of aggregation functions, separated by commas, that will be applied to each column of aggregated_columns.
	For example: MIN , MAX , SUM , AVG , COUNT , VARIANCE , and STDDEV

the output column names will be [original_column]_[aggregation]

selection_functions: argument to SQL WHERE. It can involve any column in the table and not just columns_without_aggregation and aggregated_columns.

having_functions: argument to SQL HAVING. It can involve any column in the table and the columns after aggregation in the form [original_column]_[aggregation].

sort_by: argument to SQL SORT BY

max_results: integer N to use in "LIMIT N" at the end
"""
from abc import ABC, abstractmethod

import duckdb


class LLM_Database(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def get_database_name(self):
        """This method should return the name of the database"""
        pass

    @abstractmethod
    def get_database_info(self):
        """This method should provide a comprehensive description of the
        tables in the database as they will be accessible using the
        self.sql_query method, and any specifics about those tables and columns
        """
        pass

    @abstractmethod
    def get_tables(self):
        """This method should return a list of strings. Each contains one
        valid table name that can be used in the query (see self.sql_query)
        """
        pass

    @abstractmethod
    def sql_query(self, query):
        """This method should *safely* run queries in the database.
        Make sure that queries run by this method cannot change the contents
        of the database or do anything harmful. This method should also implement
        safeguards like pre-filtering per account id.

        A usual way of doing it would be:
        WITH exposed_table AS
        (
            SELECT * FROM real_table
            WHERE real_table.account_id = <ACCOUNT_ID>
        )
        [Insert user query that uses exposed_table]
        """
        pass

    def get_full_database_description(self):
        description = [
            "Information about the tables in the dataset can be found in the <database_tables></database_tables>:"
        ]
        description = ["<database_tables>"]
        for tbl in self.get_tables():
            df = self.sql_query(f"SELECT * FROM '{tbl}' LIMIT 5")
            description.append("<database_table>")
            description.append(f"<table_name>{tbl}</table_name>")

            # information
            description.append(
                f"<table_columns>{','.join(list(df.columns))}</table_columns>"
            )
            description.append(
                f"<table_column_types>{','.join([str(x) for x in list(df.dtypes)])}</table_column_types>"
            )
            xml_sample = df.to_xml(
                index=False, xml_declaration=False, root_name=f"table_sample_data"
            )
            description.append(xml_sample)
            description.append("</database_table>")
        description.append("</database_tables>")
        description.append(
            "Note that the information in <table_sample_data></table_sample_data> is just a very small sample of the records in the table. To retrieve real data, it is necessary to actually query the table."
        )
        description.append(self.get_database_info())
        return "\n".join(description)


class SampleOrder_LLM_DB(LLM_Database):
    def __init__(self):
        self.db_path = f"gat_llm/tools/query_database_sales_data_sample.csv"

    def get_database_name(self):
        return "Sales_database"

    def get_tables(self):
        return ["tblSales"]

    def sql_query(self, query, max_desired_results=400):
        # handle when the LLM decides to use WITH
        if len(query) > 4 and query[0:4].lower() == "with":
            query = ", " + query[5:]

        # handle LIMIT
        query_lines = [
            x for x in query.replace(";", "").splitlines() if x.strip() != ""
        ]
        if "limit " not in query_lines[-1].lower():
            query = query.replace(";", "") + f"\nLIMIT {max_desired_results + 1}"
        else:
            assert (
                int(query_lines[-1].split()[-1]) <= max_desired_results
            ), f"Error: the LIMIT clause cannot request for more than {max_desired_results} results"

        q = f"""
WITH tblSales AS
(
    SELECT * FROM '{self.db_path}'
)
{query}
"""
        return duckdb.sql(q).df()

    def get_database_info(self):
        return f"In table tblSales: Column PRODUCTCODE is a unique identifier of the product. ORDERNUMBER is a unique identifier of the order."


class ToolQueryLLMDB:
    def __init__(self, LLM_Database, max_records=100):
        self.db = LLM_Database
        self.max_records = max_records
        self.name = f"query_database_{self.db.get_database_name()}"

        db_description = self.db.get_full_database_description()

        self.tool_description = {
            "name": self.name,
            "description": f"""Generates SQL code that will be used to query the database {self.db.get_database_name()}.
If the query is too complicated, involving multiple JOINs or conditions, make sure to ask the user if the results make sense.
The details of the database are as follows:

{db_description}

Never use any commands that can modify the tables or records in the database.
If the number of records exceeds {max_records}, inform that to the user and help him refine the query to narrow down the search.
If the query is executed successfully, a table in XML format containing the results will be returned.
If an error happens, the error description will be returned.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "sql_code": {
                        "type": "string",
                        "description": """Valid SQL code to query the database.""",
                    },
                },
                "required": ["sql_code"],
            },
        }

    def __call__(self, sql_code, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        try:
            ans = self.db.sql_query(sql_code, max_desired_results=self.max_records)
            if len(ans) > self.max_records:
                final_ans = [
                    "SQL code NOT executed. Too many records. Please refine the search.",
                    f"Number of records found: {len(ans)}. Maximum allowed: {self.max_records}",
                ]
            else:
                final_ans = [
                    "SQL code executed correctly. Results:",
                    ans.to_xml(
                        index=False, xml_declaration=False, root_name=f"query_results"
                    ),
                ]
        except Exception as ex:
            final_ans = ["SQL code NOT executed. Error description:", str(ex)]

        return "\n".join(final_ans)



import os
import numpy as np

rng = np.random.default_rng()


class ToolPlotWithGraphviz:
    def __init__(self):
        self.name = "plot_with_graphviz"

        self.save_code = "graph.write_png('media/graph.png')"

        self.tool_description = {
            "name": self.name,
            "description": f"""Generates a graph visualization using the provided custom Python code graph_code, whose only dependencies should be pydot. This tool is useful when the user requests graph visualizations to be generated. Prefer to use a horizontal layout and the `ortho` mode.

If the user didn't specifically ask for a graph visualization, confirm if the user wants to generate a graph visualization before calling this tool.

The last line of the graph code must save the figure to 'media/graph.png'. The last line of the code should be: {self.save_code}

The user can view the images without exposing the auto-generated file names. Do NOT include actual file names in the answer. Do NOT include the <path_to_image></path_to_image> tag in the answer. Only mention that the graph image has been generated successfully.

Raises ValueError: if the code to be executed was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "graph_code": {
                        "type": "string",
                        "description": "Python code that, when executed, will generate a graph visualization using pydot",
                    },
                },
                "required": ["graph_code"],
            },
        }

    def __call__(self, graph_code, **kwargs):
        os.makedirs("media", exist_ok=True)
        # fix weird save attempts
        graph_code = graph_code.splitlines()
        graph_code = [x for x in graph_code if not x.startswith("graph.write_png")]
        graph_code.append(self.save_code)
        graph_code = "\n".join(graph_code)

        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"'media/graph_{rng_num}.png'"
        graph_code = graph_code.replace("'media/graph.png'", target_file)

        try:
            exec(graph_code)
        except Exception as e:
            return f"Graph was NOT generated.\nError description: {str(e)}"

        target_file = target_file.replace("'", "")
        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        ans = ["<image>"]
        ans.append(f"<path_to_image>{target_file}</path_to_image>")
        ans.append("</image>")
        return "\n".join(ans)



import os

from openai import OpenAI
from .openai import LLM_GPT_OpenAI


class LLM_Grok(LLM_GPT_OpenAI):
    def __init__(self, model_size):
        """Constructor
        Arguments:
            model_size - Grok model to use
        """
        if model_size == "Grok2Vision xAI":
            self.model_id = "grok-2-vision-1212"
            self.llm_description = "Grok 2 (medium-sized LLM) - directly from xAI"
            self.price_per_M_input_tokens = 2
            self.price_per_M_output_tokens = 10
        elif model_size == "Grok4":
            self.model_id = "grok-4-0709"
            self.llm_description = "Grok 4 (medium-sized LLM) - directly from xAI"
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15

        try:
            self.openai_client = OpenAI(
                api_key=os.environ.get("GROK_API_KEY"),
                base_url="https://api.x.ai/v1",
            )
        except Exception:
            self.openai_client = None

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []



import os

import qrcode
import numpy as np

rng = np.random.default_rng()


class ToolMakeQRCode:
    def __init__(self):
        self.name = "make_qr_code"

        self.tool_description = {
            "name": self.name,
            "description": """Generates an image of a QR code given the text to be coded and the QR code configurations.

The user can view the QR code generated without exposing the auto-generated file names. Do NOT include actual file names in the answer. Do NOT include the <path_to_image></path_to_image> tag in the answer. Only mention that the QR code has been generated successfully.

Raises ValueError: if any parameter was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "qr_text": {
                        "type": "string",
                        "description": "String to be encoded in the QR code image",
                    },
                    "error_correction": {
                        "type": "string",
                        "enum": ["low", "medium", "high", "highest"],
                        "description": """Optional. If specified, must be one of the values in the enum. Lower values generate smaller QR codes but are less tolerant to faults. Higher values generate larger QR codes but the QR code is still valid in the presence of more damage to the image. If no error_correction value is provided, the default 'medium' will be used.

<correction_values>
<value>
<name>low</name>
<description>About 7% or less errors can be corrected (low).</description>
</value>
<value>
<name>medium</name>
<description>Default. About 15% or less errors can be corrected (medium).</description>
</value>
<value>
<name>high</name>
<description>About 25% or less errors can be corrected (high).</description>
</value>
<value>
<name>highest</name>
<description>About 30% or less errors can be corrected (highest).</description>
</value>
</correction_values>""",
                    },
                    "box_size": {
                        "type": "integer",
                        "description": """Optional. If specified, controls how many pixels each "box" of the QR code has. The size of the image increases for larger values.
The default value is 10.
Only use values from 5 (small image) to 50 (very large image).""",
                    },
                },
                "required": ["qr_text"],
            },
        }

    def __call__(self, qr_text, error_correction="medium", box_size=10, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        error_corrections = {
            "low": qrcode.constants.ERROR_CORRECT_L,
            "medium": qrcode.constants.ERROR_CORRECT_M,
            "high": qrcode.constants.ERROR_CORRECT_Q,
            "highest": qrcode.constants.ERROR_CORRECT_H,
        }
        if error_correction not in error_corrections.keys():
            return f"error_correction must be one of {error_corrections.keys()}"

        qr = qrcode.QRCode(
            version=1,
            error_correction=error_corrections[error_correction],
            box_size=10,
            border=4,
        )
        qr.add_data(qr_text)
        qr.make(fit=True)

        img = qr.make_image(fill_color="black", back_color="white")

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/qr_{rng_num}.png"
        img.save(target_file)

        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        # with open(f'plot_code_{rng_num}.txt', 'w', encoding='utf-8') as debug_file:
        #    p_str = plot_code + f'\n***{target_file} {os.path.isfile(target_file)}'
        #    debug_file.write(p_str)

        ans = ["<image>"]
        ans.append(f"<path_to_image>{target_file}</path_to_image>")
        ans.append("</image>")
        return "\n".join(ans)



""" Set of available and useful LLMs (mostly posted on AWS Bedrock)
"""
import warnings

# outdated
from .llm_providers.aws_bedrock import LLM_Llama13b
from .llm_providers.aws_bedrock import LLM_Llama70b
from .llm_providers.aws_bedrock import LLM_Llama3
from .llm_providers.aws_bedrock import LLM_Claude2_1_Bedrock
from .llm_providers.aws_bedrock import LLM_Claude_Instant_1_2_Bedrock

# current
from .llm_providers.openai import LLM_GPT_OpenAI
from .llm_providers.anthropic import LLM_Claude_Anthropic
from .llm_providers.aws_bedrock import LLM_Claude_Bedrock
from .llm_providers.aws_bedrock import LLM_Mistral_Bedrock
from .llm_providers.aws_bedrock_cohere import LLM_Command_Cohere
from .llm_providers.maritaca import LLM_Maritalk
from .llm_providers.aws_bedrock_nova import LLM_Nova_Bedrock
from .llm_providers.deepseek import LLM_Deepseek
from .llm_providers.grok import LLM_Grok

# local
from .llm_providers.ollama import LLM_Ollama


warnings.simplefilter("always", DeprecationWarning)


class LLM_Provider:
    outdated_llms = [
        "Claude 3.5 Sonnet - Anthropic",
        "Claude 3.5 Sonnet - Bedrock",
        "Claude 3 Opus - Anthropic",
        "Claude 3 Haiku - Anthropic",
        "Claude 3 Haiku - Bedrock",
        "Claude 3 Sonnet - Bedrock",
        "Claude 3 Opus - Bedrock",
        "Claude 2.1",
        "Claude Instant 1.2",
        "Llama2 13b",
        "Llama2 70b",
        "Llama3 8b instruct",
        "Llama3 70b instruct",
        "Grok2Vision - Grok",
    ]

    allowed_llms = [
        # Local
        "Qwen 3 8b - Ollama",
        "Qwen 3 14b - Ollama",
        "Llama4 16x17b - Ollama",
        "Qwen 2.5vl 7b - Ollama",
        "DeepSeek R1 14b - Ollama",
        # AWS
        "Amazon Nova Micro 1.0 - Bedrock",
        "Amazon Nova Lite 1.0 - Bedrock",
        "Amazon Nova Pro 1.0 - Bedrock",
        # Grok
        "Grok4 - Grok",
        # Maritaca
        "Sabia3 - Maritaca",
        # OpenAI
        "GPT 4.1 - OpenAI",
        "GPT 4o - OpenAI",
        "GPT 3.5 - OpenAI",
        "GPT 4o mini - OpenAI",
        # Anthropic
        "Claude 4 Opus - Anthropic",
        "Claude 4 Sonnet - Anthropic",
        "Claude 4 Opus - Bedrock",
        "Claude 4 Sonnet - Bedrock",
        "Claude 3.7 Sonnet - Anthropic",
        "Claude 3.5 Haiku - Anthropic",
        "Claude 3.7 Sonnet - Bedrock",
        "Claude 3.5 Haiku - Bedrock",
        # Misc
        "DeepSeekV3 Chat - DeepSeek",
        "Command R - Bedrock",
        "Command RPlus - Bedrock",
        "Mistral Mixtral 8x7B",
        "Mistral Large v1",
        "Llama3_1 8b instruct",
        "Llama3_1 70b instruct",
        "Llama3_1 405b instruct",
    ]

    def get_llm(bedrock_client, llm):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
            llm - which LLM to use. Check LLM_Service.allowed_llms for a list
        """
        if llm in LLM_Provider.outdated_llms:
            warn_msg = f"Selected model is outdated: {llm}. Consider switching to a newer model."
            warnings.warn(warn_msg, DeprecationWarning)

        assert (
            llm in LLM_Provider.allowed_llms + LLM_Provider.outdated_llms
        ), f"LLM has to be one of {LLM_Provider.allowed_llms}"
        if llm == "Claude 2.1":
            return LLM_Claude2_1_Bedrock(bedrock_client)
        elif llm == "Claude Instant 1.2":
            return LLM_Claude_Instant_1_2_Bedrock(bedrock_client)
        elif llm == "Llama2 13b":
            return LLM_Llama13b(bedrock_client)
        elif llm == "Llama2 70b":
            return LLM_Llama70b(bedrock_client)

        # Local - Ollama
        elif llm == "Llama4 16x17b - Ollama":
            return LLM_Ollama(model="Llama4 16x17b Ollama")
        elif llm == "Qwen 3 8b - Ollama":
            return LLM_Ollama(model="Qwen 3 8b Ollama")
        elif llm == "Qwen 3 14b - Ollama":
            return LLM_Ollama(model="Qwen 3 14b Ollama")
        elif llm == "Qwen 2.5vl 7b - Ollama":
            return LLM_Ollama(model="Qwen 2.5vl 7b Ollama")
        elif llm == "DeepSeek R1 14b - Ollama":
            return LLM_Ollama(model="DeepSeek R1 14b Ollama")

        # Amazon
        elif llm == "Amazon Nova Micro 1.0 - Bedrock":
            return LLM_Nova_Bedrock(bedrock_client, model_size="Nova_Micro")
        elif llm == "Amazon Nova Lite 1.0 - Bedrock":
            return LLM_Nova_Bedrock(bedrock_client, model_size="Nova_Lite")
        elif llm == "Amazon Nova Pro 1.0 - Bedrock":
            return LLM_Nova_Bedrock(bedrock_client, model_size="Nova_Pro")

        # Maritaca
        elif llm == "Sabia3 - Maritaca":
            return LLM_Maritalk(model_size="Sabia3 Maritaca")

        # Grok
        elif llm == "Grok2Vision - Grok":
            return LLM_Grok(model_size="Grok2Vision xAI")
        elif llm == "Grok4 - Grok":
            return LLM_Grok(model_size="Grok4")

        # DeepSeek
        elif llm == "DeepSeekV3 Chat - DeepSeek":
            return LLM_Deepseek(model_size="Deepseek Chat")
        elif llm == "DeepSeekR1 Reasoner - DeepSeek":
            return LLM_Deepseek(model_size="Deepseek Reasoner")

        # Llama 3.1 family
        elif llm == "Llama3_1 8b instruct":
            return LLM_Llama3(bedrock_client, model="Llama3_1 8B Instruct - Bedrock")
        elif llm == "Llama3_1 70b instruct":
            return LLM_Llama3(bedrock_client, model="Llama3_1 70B Instruct - Bedrock")
        elif llm == "Llama3_1 405b instruct":
            return LLM_Llama3(bedrock_client, model="Llama3_1 405B Instruct - Bedrock")

        elif llm == "Llama3 8b instruct":
            return LLM_Llama3(bedrock_client, model="Llama3 8B Instruct - Bedrock")
        elif llm == "Llama3 70b instruct":
            return LLM_Llama3(bedrock_client, model="Llama3 70B Instruct - Bedrock")
        elif llm == "Command R - Bedrock":
            return LLM_Command_Cohere(bedrock_client, model_size="Command R Cohere 1")
        elif llm == "Command RPlus - Bedrock":
            return LLM_Command_Cohere(
                bedrock_client, model_size="Command RPlus Cohere 1"
            )
        elif llm == "Mistral Mixtral 8x7B":
            return LLM_Mistral_Bedrock(bedrock_client, model_size="Mixtral 8x7B v0:1")
        elif llm == "Mistral Large v1":
            return LLM_Mistral_Bedrock(bedrock_client, model_size="Mistral Large v1")

        # Legacy Claude
        elif llm == "Claude 3 Opus - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Opus")
        elif llm == "Claude 3 Sonnet - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Sonnet")
        elif llm == "Claude 3 Haiku - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Haiku")
        elif llm == "Claude 3 Opus - Anthropic":
            return LLM_Claude_Anthropic(model_size="Opus 3 Anthropic")
        elif llm == "Claude 3 Haiku - Anthropic":
            return LLM_Claude_Anthropic(model_size="Haiku 3 Anthropic")

        # current Claude
        elif llm == "Claude 4 Sonnet - Anthropic":
            return LLM_Claude_Anthropic(model_size="Sonnet 4 Anthropic")
        elif llm == "Claude 4 Opus - Anthropic":
            return LLM_Claude_Anthropic(model_size="Opus 4 Anthropic")
        elif llm == "Claude 4 Sonnet - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Sonnet 4")
        elif llm == "Claude 4 Opus - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Opus 4")
        elif llm == "Claude 3.7 Sonnet - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Sonnet 3.7")
        elif llm == "Claude 3.5 Sonnet - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Sonnet 3.5")
        elif llm == "Claude 3.5 Haiku - Bedrock":
            return LLM_Claude_Bedrock(bedrock_client, model_size="Haiku 3.5")
        elif llm == "Claude 3.5 Sonnet - Anthropic":
            return LLM_Claude_Anthropic(model_size="Sonnet 3.5 Anthropic")
        elif llm == "Claude 3.7 Sonnet - Anthropic":
            return LLM_Claude_Anthropic(model_size="Sonnet 3.7 Anthropic")
        elif llm == "Claude 3.5 Haiku - Anthropic":
            return LLM_Claude_Anthropic(model_size="Haiku 3.5 Anthropic")

        # OpenAI
        elif llm == "GPT 4.1 - OpenAI":
            return LLM_GPT_OpenAI(model_size="GPT4_1 OpenAI")
        elif llm == "GPT 3.5 - OpenAI":
            return LLM_GPT_OpenAI(model_size="GPT3_5 OpenAI")
        elif llm == "GPT 4o - OpenAI":
            return LLM_GPT_OpenAI(model_size="GPT4o OpenAI")
        elif llm == "GPT 4o mini - OpenAI":
            return LLM_GPT_OpenAI(model_size="GPT4o mini OpenAI")



import os
import re
import pypdf
import base64
import requests
from pathlib import Path
from io import BytesIO

import pandas as pd
from markitdown import MarkItDown
from pdf2image import convert_from_path


def extract_text(file) -> str:
    file = Path(file)
    extension = file.suffix.lower()

    # Check the file type and read
    if extension in [".txt", ".py", ".md", ".srt", ".js", ".jsx", ".html", ".css"]:
        with open(file, "r", encoding="utf-8") as f:
            ans = f.read()
        ans = f"<contents>\n{ans}\n</contents>"
    elif extension in [".docx", ".pptx", ".xlsx", ".xls"]:
        md = MarkItDown()
        ans = md.convert(file).text_content
    elif extension == ".pdf":
        ans = pdf_to_xml(file)
    elif extension == ".csv":
        ans = csv_to_xml(file)
    else:
        raise ValueError(f"Unsupported file extension: {extension}")

    return ans


def pdf_to_xml(pdf_path):
    # Open the PDF
    pdf = pypdf.PdfReader(pdf_path)
    xml_content = ["<contents>"]

    # Loop through pages and extract text
    for page_num, page in enumerate(pdf.pages, start=1):
        text = page.extract_text()
        xml_content.append(f"<page_{page_num}>{text}</page_{page_num}>")

    xml_content.append("</contents>")
    return "\n".join(xml_content)


def pdf_pages_to_base64_images(pdf_path, dpi=200, fmt="JPEG", jpeg_quality=95):
    """
    Convert each page of a PDF into a Base64-encoded image.

    Args:
        pdf_path (str): Path to the PDF file.
        dpi (int): Resolution of output images (higher dpi = better quality).
        fmt (str): Image format, e.g., 'PNG', 'JPEG'.

    Returns:
        List[str]: Base64-encoded strings of each page image.
    """
    images = convert_from_path(pdf_path, dpi=dpi)
    base64_images = []

    for img in images:
        buffered = BytesIO()
        img.save(buffered, format=fmt, quality=jpeg_quality)
        img_base64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
        base64_images.append(img_base64)

    return base64_images


def sanitize_column_name(name):
    """
    Convert column names to valid XML tag names by:
    - Removing or replacing invalid characters
    - Replacing whitespace with underscores
    - Removing carriage return and newline characters
    """
    name = name.strip().replace("\r", "").replace("\n", "")
    name = re.sub(r"[^a-zA-Z0-9_\-]", "_", name)
    if name[0].isdigit():
        name = f"col_{name}"
    return name


def csv_to_xml(csv_path):
    # Try different encodings and delimiter if necessary
    for encoding in ["utf-8", "latin-1", "utf-16"]:
        try:
            df = pd.read_csv(csv_path, encoding=encoding)
            break
        except UnicodeDecodeError:
            continue
        except pd.errors.ParserError:
            # Try with semicolon as delimiter if parsing fails
            try:
                df = pd.read_csv(csv_path, encoding=encoding, delimiter=";")
                break
            except Exception:
                continue
    else:
        raise ValueError(
            "Unable to read the CSV file with common encodings and delimiters."
        )

    # Remove unnamed columns and sanitize column names
    df = df.loc[:, ~df.columns.str.startswith("Unnamed:")]
    df.columns = [sanitize_column_name(col) for col in df.columns]

    # Convert DataFrame to XML
    xml_data = df.to_xml(index=False, root_name="data", row_name="record")
    return xml_data


def normalize_xml_content(xml_content):
    # Split into lines, strip trailing whitespace from each, then rejoin
    xml_content = "\n".join(line.rstrip() for line in xml_content.splitlines())

    # Add EOF
    xml_content += "\n"
    return xml_content


class ToolReadLocalFile:
    def __init__(self, query_llm=None):
        self.name = "read_local_files"
        self.query_llm = query_llm

        self.tool_description = {
            "name": self.name,
            "description": """Reads one or more files and return their contents. Provide one file per line.
Only read files of type <allowed_extensions></allowed_extensions>:
<allowed_extensions>
pdf
docx
txt
md
pptx
xlsx
xls
py
srt
csv
</allowed_extensions>

Do not attempt to read files outside the types described in the <allowed_extensions></allowed_extensions>.
Do not attempt to read files that are usually in binary format.

Raises ValueError: if the file does not exist.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "path_to_files": {
                        "type": "string",
                        "description": """Local path to the files whose contents should be retrieved. Provide one file per line as in the <example></example>:
<example>
file1.pdf
file2.docx
subfolder/file3.txt
</example>""",
                    },
                    "prompt": {
                        "type": "string",
                        "description": "Prompt that can be used to directly query and retrieve information from the files, without fetching their full content.",
                    },
                    "pdfs_to_read_as_images": {
                        "type": "string",
                        "description": """Local path of PDF files that need to be read as images.
The following <use_case_examples></use_case_examples> provide some examples of use cases that require analyzing the PDF as images:
<use_case_examples>
<use_case_example>
Automated proofreading: Identify graphs, diagrams, or handwritten texts drawn.
</use_case_example>
<use_case_example>
Recognition of mathematical formulas: Extract handwritten equations or scientific diagrams.
</use_case_example>
<use_case_example>
Analysis of academic works: Check figures, tables, and graphs.
</use_case_example>
</use_case_examples>
Provide this parameter only after analyzing whether there is a need to read images from the PDF, based on the user's query. This is important because parsing images is expensive and we do not want to do it unnecessarily.

Provide one file per line as in the <paths_example></paths_example>:
<paths_example>
file1.pdf
subfolder/file2.pdf
</paths_example>""",
                    },
                },
                "required": ["path_to_files", "prompt"],
            },
        }

    def __call__(self, path_to_files, prompt="", pdfs_to_read_as_images="", **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        # make sure all pdfs to read as images are in the file list
        for x in pdfs_to_read_as_images:
            if x not in path_to_files:
                path_to_files.append(x)

        final_ans = ["<files>"]
        all_files = [x.strip() for x in path_to_files.splitlines() if x.strip() != ""]
        pdfs_to_read_as_images = [
            x.strip() for x in pdfs_to_read_as_images.splitlines() if x.strip() != ""
        ]
        b64_images = None

        for path_to_file in all_files:
            ans = ""
            if not os.path.isfile(path_to_file):
                ans = f"Error: Did not find file `{path_to_file}`"
                ans = f"<error>\n{ans}\n</error>"

            if path_to_file.startswith("http://") or path_to_file.startswith(
                "https://"
            ):
                try:
                    response = requests.get(path_to_file)
                    response.raise_for_status()
                    with open("/tmp/temp_downloaded_file.txt", "wb") as f:
                        f.write(response.content)
                    ans = extract_text("/tmp/temp_downloaded_file.txt")
                    if path_to_file in pdfs_to_read_as_images:
                        b64_images = pdf_pages_to_base64_images(path_to_file)
                    os.remove("/tmp/temp_downloaded_file.txt")
                except Exception as e:
                    ans = f"Error: Failed to fetch or process URL `{path_to_file}`: {str(e)}"
                    ans = f"<error>\n{ans}\n</error>"
            else:
                if not os.path.isfile(path_to_file):
                    ans = f"Error: Did not find file `{path_to_file}`"
                    ans = f"<error>\n{ans}\n</error>"
                else:
                    try:
                        ans = extract_text(path_to_file)
                        if path_to_file in pdfs_to_read_as_images:
                            b64_images = pdf_pages_to_base64_images(path_to_file)
                    except Exception as e:
                        ans = f"Error: Failed to process the file `{path_to_file}`: {str(e)}"
                        ans = f"<error>\n{ans}\n</error>"

            final_ans.append("<file>")
            final_ans.append(f"<file_name>{path_to_file}</file_name>")
            final_ans.append(ans)
            final_ans.append("</file>")

        final_ans.append("</files>")
        final_ans = "\n".join(final_ans)

        # if a subquery has been asked
        if prompt is not None and prompt != "":
            if self.query_llm is None:
                return "Error: Cannot retrieve data from the document because a LLM has not been provided. Please set prompt to '' to return the full document."
            sys_prompt = "Read the contents of the following <files></files> to answer questions:\n"
            sys_prompt += final_ans
            llm_ans = self.query_llm(
                prompt,
                b64images=b64_images,
                system_prompt=sys_prompt,
            )
            for x in llm_ans:
                pass
            final_ans = x

        final_ans = normalize_xml_content(final_ans)

        return final_ans



import re
import json
import time

import anthropic
from .base_service import LLM_Service


class LLM_Claude_Anthropic(LLM_Service):
    def __init__(self, model_size, use_caching=True):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        self.use_caching = use_caching
        self.anthropic_client = anthropic.Anthropic()
        if model_size == "Opus 4 Anthropic":
            self.model_id = "claude-opus-4-20250514"
            self.llm_description = (
                "Anthropic Claude 4 Opus (Large-size LLM) - directly from Anthropic"
            )
            self.price_per_M_input_tokens = 15
            self.price_per_M_output_tokens = 75
        elif model_size == "Sonnet 4 Anthropic":
            self.model_id = "claude-sonnet-4-20250514"
            self.llm_description = (
                "Anthropic Claude 4 Sonnet (Medium-size LLM) - directly from Anthropic"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Sonnet 3.7 Anthropic":
            self.model_id = "claude-3-7-sonnet-20250219"
            self.llm_description = "Anthropic Claude 3.7 Sonnet (Medium-size LLM) - directly from Anthropic"
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Sonnet 3.5 Anthropic":
            self.model_id = "claude-3-5-sonnet-20241022"
            self.llm_description = "Anthropic Claude 3.5 Sonnet (Medium-size LLM) - directly from Anthropic"
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Haiku 3.5 Anthropic":
            self.model_id = "claude-3-5-haiku-20241022"
            self.llm_description = (
                "Anthropic Claude 3.5 Haiku (Small-size LLM) - directly from Anthropic"
            )
            self.price_per_M_input_tokens = 1
            self.price_per_M_output_tokens = 5
        elif model_size == "Opus 3 Anthropic":
            self.model_id = "claude-3-opus-20240229"
            self.llm_description = (
                "Anthropic Claude 3 Opus (Large-size LLM) - directly from Anthropic"
            )
            self.price_per_M_input_tokens = 15
            self.price_per_M_output_tokens = 75
        elif model_size == "Haiku 3 Anthropic":
            self.model_id = "claude-3-haiku-20240307"
            self.llm_description = (
                "Anthropic Claude 3 Haiku (Small-size LLM) - directly from Anthropic"
            )
            self.price_per_M_input_tokens = 0.25
            self.price_per_M_output_tokens = 1.25

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop_sequences": [],  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        ans = {}
        if msg_list[0]["role"] == "system":
            ans["system"] = msg_list[0]["content"]
        ans["messages"] = []
        for msg in msg_list[1:]:
            ans["messages"].append(msg)
        return ans

    def invoke_streaming(
        self,
        prompt,
        b64image=None,
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=5,
        cur_fail_sleep=6,
    ):
        """
        Invokes the Claude 3 model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered.
            In the case of Claude 3, it is a dictionary with keys
            system -> str, messages -> list
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        tools: description of the tools that can be used, Claude format
        tool_invoker_fn: function that invokes the tools. Arguments are:
                function name - function to call
                return_results_only - we set to True because we already use Claude format
                kwargs - arguments to the tool that will be called
        max_retries: how many attempts to call the model
        cur_fail_sleep: how long to wait between model calls (this gets incremented)
        :return: Inference response from the model.
        """
        # Messages that had to be added because of function use
        self.tool_use_added_msgs = []

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences

        body["system"] = prompt["system"]
        body["messages"] = prompt["messages"].copy()

        # apply caching to System Prompt
        if self.use_caching:
            if isinstance(body["system"], str):
                body["system"] = [
                    {
                        "type": "text",
                        "text": body["system"],
                        "cache_control": {"type": "ephemeral"},
                    }
                ]
            elif isinstance(body["system"], list):
                body["system"][-1]["cache_control"] = {"type": "ephemeral"}

        if tools is None:
            body["messages"].append({"role": "assistant", "content": postpend})
        else:
            if self.use_caching:
                # if caching, append the caching structure to the last tool
                tools[-1]["cache_control"] = {"type": "ephemeral"}
            body["tools"] = tools

            assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

        for k in range(max_retries):
            try:
                self.debug_body = body
                llm_body_changed = True
                while llm_body_changed:
                    llm_body_changed = False
                    response = self.anthropic_client.messages.create(**body)

                    word_count = len(re.findall(r"\w+", str(body["messages"])))
                    print(f"Invoking {self.llm_description}. Word count: {word_count}")

                    # stream responses
                    partial_ans = self._response_gen(response, postpend)
                    x = ""
                    for x in partial_ans:
                        yield x
                    cur_ans = x

                    if self.cur_tool_spec is not None:
                        # tool use has been required. Let's do it
                        # TODO: update upstream to reflect the inclusion of a response
                        # TODO: probably rework gradio UI to re-instantiate things every chat, or keep an instance per chat ID
                        tool_ans = tool_invoker_fn(
                            self.cur_tool_spec["name"],
                            return_results_only=True,
                            **self.cur_tool_spec["input"],
                        )

                        # append assistant responses
                        assistant_msg = {"role": "assistant", "content": []}
                        if cur_ans is not None and cur_ans.strip() != "":
                            assistant_msg["content"].append(
                                {
                                    "type": "text",
                                    "text": cur_ans,
                                },
                            )
                        assistant_msg["content"].append(self.cur_tool_spec)

                        next_user_msg = {
                            "role": "user",
                            "content": [
                                {
                                    "type": "tool_result",
                                    "tool_use_id": self.cur_tool_spec["id"],
                                    "content": tool_ans,
                                }
                            ],
                        }
                        body["messages"].append(assistant_msg)
                        body["messages"].append(next_user_msg)

                        # keep a log of messages that had to be appended due to tool use
                        self.tool_use_added_msgs.append(assistant_msg)
                        self.tool_use_added_msgs.append(next_user_msg)
                        llm_body_changed = True

                    # TODO: Include proper token count and pricing
                    ans_word_count = len(
                        re.findall(r"\w+", postpend + cur_ans + str(self.stop_reason))
                    )
                    self.word_counts.append(
                        {
                            "request_word_count": word_count,
                            "answer_word_count": ans_word_count,
                            "price_estimate": 0.001
                            * (0.003 * word_count + 0.015 * ans_word_count),
                            "exec_time_in_s": time.time() - t0,
                        }
                    )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."

    def _response_gen(self, response_body, postpend=""):
        cur_ans = ""
        cur_tool_spec = None
        for x in response_body:
            txt = ""
            if hasattr(x, "type") and x.type == "message_start":
                print(x)
            if hasattr(x, "content_block"):
                if x.content_block.type == "text":
                    txt = x.content_block.text
                elif x.content_block.type == "tool_use":
                    cur_tool_spec = x.content_block.__dict__.copy()
                    cur_tool_spec["input"] = ""

            elif hasattr(x, "delta"):
                txt = x.delta.text if hasattr(x.delta, "text") else ""
                if cur_tool_spec is not None:
                    cur_tool_spec["input"] += (
                        x.delta.partial_json if hasattr(x.delta, "partial_json") else ""
                    )

            if txt != "":
                cur_ans += txt
                yield postpend + cur_ans

            stop_reason = (
                x.delta.stop_reason
                if hasattr(x, "delta") and hasattr(x.delta, "stop_reason")
                else None
            )
            if stop_reason is not None and stop_reason == "stop_sequence":
                stop_txt = x.delta.stop_sequence
                yield postpend + cur_ans + stop_txt
                break
        if cur_tool_spec is not None:
            cur_tool_spec["input"] = json.loads(cur_tool_spec["input"])
        self.cur_tool_spec = cur_tool_spec
        self.stop_reason = stop_reason



import pytest
from unittest.mock import patch
from gat_llm.tools.read_file_names_in_local_folder import ToolReadLocalFolder


@pytest.fixture
def mock_os_path_isdir():
    with patch("os.path.isdir") as mock:
        mock.return_value = True
        yield mock


@pytest.fixture
def mock_os_listdir():
    with patch("os.listdir") as mock:
        mock.return_value = ["file1.txt", "file2.py", "folder"]
        yield mock


@pytest.fixture
def mock_os_path_isfile():
    with patch("os.path.isfile") as mock:
        mock.side_effect = lambda x: not x.endswith("folder")
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    trlf = ToolReadLocalFolder()
    ans = trlf("/path/to/folder", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_read_file_names_success(
    mock_os_path_isdir, mock_os_listdir, mock_os_path_isfile
):
    trlf = ToolReadLocalFolder()
    result = trlf("/path/to/folder")
    assert "<files>" in result
    assert "file1.txt</file>" in result
    assert "file2.py</file>" in result
    assert "</files>" in result


@patch("os.path.isdir", return_value=False)
def test_read_file_names_invalid_folder(mock_isdir):
    trlf = ToolReadLocalFolder()
    result = trlf("/invalid/folder")
    assert "Error: Did not find folder" in result


# Add more tests for different scenarios and edge cases



import json

from unittest.mock import MagicMock
from unittest.mock import Mock

import pytest

from gat_llm.llm_invoker import LLM_Provider


@pytest.mark.parametrize(
    "outdated_llm",
    [
        ("Claude 3 Opus - Anthropic"),
        ("Claude 3 Haiku - Anthropic"),
        ("Claude 3 Haiku - Bedrock"),
        ("Claude 3 Sonnet - Bedrock"),
        ("Claude 3 Opus - Bedrock"),
        ("Claude 2.1"),
        ("Claude Instant 1.2"),
        ("Llama2 13b"),
        ("Llama2 70b"),
        ("Llama3 8b instruct"),
        ("Llama3 70b instruct"),
    ],
)
def test_outdated_llm_warning(outdated_llm):
    # Ensures that outdated models show warnings
    with pytest.deprecated_call():
        bedrock_client = None
        LLM_Provider.get_llm(bedrock_client, outdated_llm)


@pytest.mark.parametrize("llm", LLM_Provider.allowed_llms + LLM_Provider.outdated_llms)
def test_model_creation(llm):
    # Checks if the models are created successfully
    # (even without the API key)
    bedrock_client = None
    LLM_Provider.get_llm(bedrock_client, llm)


def dummy_response_gen(ret_val):
    def resp_gen_func(text, postpend=""):
        # just mocks the yield part of the response
        if isinstance(text, str):
            yield text
        else:
            yield ret_val

    return resp_gen_func


@pytest.mark.parametrize(
    "llm_name", LLM_Provider.allowed_llms + LLM_Provider.outdated_llms
)
@pytest.mark.parametrize(
    "ret_val",
    [
        "Dummy LLM generation",
        # Test cases when the LLM returns nothing (e.g. pure tool use)
        "",
    ],
)
def test_if_llm_responds(llm_name, ret_val):
    # Checks if the llms actually produce an answer when required
    bedrock_client = None
    llm = LLM_Provider.get_llm(bedrock_client, llm_name)

    # mocks the clients
    llm.anthropic_client = MagicMock()
    llm.bedrock_client = MagicMock()
    llm.openai_client = MagicMock()

    # mocks the invoke methods
    llm.anthropic_client.messages.create = Mock(return_value=ret_val)
    llm.openai_client.chat.completions.create = Mock(return_value=ret_val)
    llm.bedrock_client.invoke_model_with_response_stream = Mock(
        return_value={
            "body": [
                {
                    "chunk": {
                        "bytes": json.dumps(
                            {
                                "generation": ret_val,
                                "outputs": [{"text": ret_val}],
                                "completion": ret_val,
                                "stop": "",
                            }
                        )
                    }
                }
            ]
        }
    )

    llm.cur_tool_spec = None
    llm.stop_reason = None
    llm._response_gen = dummy_response_gen(ret_val)
    ans = llm("Dummy message")
    for x in ans:
        pass

    assert x == ret_val


@pytest.mark.parametrize(
    "llm_name", LLM_Provider.allowed_llms + LLM_Provider.outdated_llms
)
def test_if_llm_exits_gracefully(llm_name):
    # Checks if the llms exit gracefully, i.e., don't raise an error
    # in this case they all should raise an error because there is no valid token
    bedrock_client = None
    llm = LLM_Provider.get_llm(bedrock_client, llm_name)

    ans = llm("Dummy message", max_retries=1, cur_fail_sleep=0)
    for x in ans:
        pass

    assert x == "Could not invoke the AI model."



import os
import subprocess


class ToolSelectVideoFrames:
    def __init__(self):
        self.name = "select_video_frames"

        self.tool_summary = f"""<tool_summary>
<tool_name>{self.name}</tool_name>
<summary>Selects relevant video frames from a video file.</summary>
</tool_summary>"""

        self.tool_description = {
            "name": self.name,
            "description": """Extracts desired video frames from a video. Use this tool when the user asks for a video to be analyzed and the answer requires the analysis of frames extracted from the video. Afterwards, you are allowed to use tools that directly analyze images.

Raises ValueError: if one of the parameters is invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "video_file_path": {
                        "type": "string",
                        "description": "Path to the video file",
                    },
                    "desired_frame_times": {
                        "type": "string",
                        "description": """Desired times from where to extract frames in Hours:Minutes:Seconds. Separate each frame using commas. Example: "00:23:45, 01:43:54" (this will be used as an argument to ffmpeg -ss [time])""",
                    },
                },
                "required": ["video_file_path", "desired_frame_times"],
            },
        }

        self.tool_summary = self.tool_description

    def __call__(
        self, video_file_path, desired_frame_times, output_folder="media", **kwargs
    ):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        if not os.path.isfile(video_file_path):
            ans = f"Error: Did not find file `{video_file_path}`"
            ans = f"<error>\n{ans}\n</error>"
            return ans

        list_frame_times = [x.strip() for x in desired_frame_times.split(",")]

        ans = ["Frames extracted:"]
        for t in list_frame_times:
            file_name = f"""frame_{t.replace(":", "_")}.jpg"""
            file_name = os.path.join(output_folder, file_name)

            all_args = (
                f"ffmpeg -y -ss {t} -i {video_file_path} -frames:v 1 -q:v 2 {file_name}"
            )
            subprocess.run(all_args, capture_output=True, text=True, shell=True)

            file_name = f"<frame>{file_name}</frame>"
            ans.append(file_name)

        return "\n".join(ans)



import re
import json
import time

from typing import Dict, List
from .base_service import LLM_Service


class LLM_Claude_Bedrock(LLM_Service):
    def __init__(self, bedrock_client, model_size, use_caching=True):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
            model_size - type of model
            use_caching - optimize pricing and speed by using caching for prompt and tools
        """
        # Only 2 Claude models from Bedrock support caching at the moment
        if model_size in ["Sonnet 3.7", "Haiku 3.5"]:
            self.use_caching = use_caching
        else:
            self.use_caching = False

        if model_size == "Sonnet 4":
            self.model_id = "us.anthropic.claude-sonnet-4-20250514-v1:0"
            self.llm_description = (
                "Anthropic Claude 4 Sonnet from AWS Bedrock (Medium-size LLM)"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Opus 4":
            self.model_id = "us.anthropic.claude-opus-4-20250514-v1:0"
            self.llm_description = (
                "Anthropic Claude 4 Opus from AWS Bedrock (Large-size LLM)"
            )
            self.price_per_M_input_tokens = 15
            self.price_per_M_output_tokens = 75
        elif model_size == "Opus":
            self.model_id = "anthropic.claude-3-opus-20240229-v1:0"
            self.llm_description = (
                "Anthropic Claude 3.0 Opus from AWS Bedrock (Large LLM)"
            )
            self.price_per_M_input_tokens = 15
            self.price_per_M_output_tokens = 75
        elif model_size == "Sonnet":
            self.model_id = "anthropic.claude-3-sonnet-20240229-v1:0"
            self.llm_description = (
                "Anthropic Claude 3.0 Sonnet from AWS Bedrock (Medium-size LLM)"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Sonnet 3.5":
            self.model_id = "anthropic.claude-3-5-sonnet-20241022-v2:0"
            self.llm_description = (
                "Anthropic Claude 3.5 Sonnet from AWS Bedrock (Medium-size LLM)"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Sonnet 3.7":
            self.model_id = "us.anthropic.claude-3-7-sonnet-20250219-v1:0"
            self.llm_description = (
                "Anthropic Claude 3.7 Sonnet from AWS Bedrock (Medium-size LLM)"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15
        elif model_size == "Haiku 3.5":
            self.model_id = "anthropic.claude-3-5-haiku-20241022-v1:0"
            self.llm_description = (
                "Anthropic Claude 3.5 Haiku from AWS Bedrock (Small-size LLM)"
            )
            self.price_per_M_input_tokens = 1
            self.price_per_M_output_tokens = 5
        elif model_size == "Haiku":
            self.model_id = "anthropic.claude-3-haiku-20240307-v1:0"
            self.llm_description = (
                "Anthropic Claude 3.0 Haiku from AWS Bedrock (Small-size LLM)"
            )
            self.price_per_M_input_tokens = 0.25
            self.price_per_M_output_tokens = 1.25

        self.bedrock_client = bedrock_client
        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            # "top_k": 250,
            # "top_p": 1,
            "stop_sequences": [],  # the regular is already implemented
            "anthropic_version": "bedrock-2023-05-31",
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        ans = {}
        if msg_list[0]["role"] == "system":
            ans["system"] = msg_list[0]["content"]
        ans["messages"] = []
        for msg in msg_list[1:]:
            ans["messages"].append(msg)
        return ans

    def invoke_streaming(
        self,
        prompt,
        b64image=None,
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=25,
        cur_fail_sleep=60,
    ):
        """
        Invokes the Claude 3 model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered.
            In the case of Claude 3, it is a dictionary with keys
            system -> str, messages -> list
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        tools: description of the tools that can be used, Claude format
        tool_invoker_fn: function that invokes the tools. Arguments are:
                function name - function to call
                return_results_only - we set to True because we already use Claude format
                kwargs - arguments to the tool that will be called
        :return: Inference response from the model.
        """
        # Messages that had to be added because of function use
        self.tool_use_added_msgs = []

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences

        body["system"] = prompt["system"]
        body["messages"] = prompt["messages"].copy()

        # apply caching to System Prompt
        if self.use_caching:
            if isinstance(body["system"], str):
                body["system"] = [
                    {
                        "type": "text",
                        "text": body["system"],
                        "cache_control": {"type": "ephemeral"},
                    }
                ]
            elif isinstance(body["system"], list):
                body["system"][-1]["cache_control"] = {"type": "ephemeral"}

        if tools is None:
            body["messages"].append({"role": "assistant", "content": postpend})
        else:
            if self.use_caching:
                # if caching, append the caching structure to the last tool
                tools[-1]["cache_control"] = {"type": "ephemeral"}
            body["tools"] = tools
            assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

        for k in range(max_retries):
            try:
                llm_body_changed = True
                while llm_body_changed:
                    llm_body_changed = False
                    response = self.bedrock_client.invoke_model_with_response_stream(
                        modelId=self.model_id,
                        body=json.dumps(body),
                    )
                    word_count = len(re.findall(r"\w+", str(body["messages"])))
                    print(f"Invoking {self.llm_description}. Word count: {word_count}")

                    # stream responses
                    partial_ans = self._response_gen(response["body"], postpend)
                    x = ""
                    for x in partial_ans:
                        yield x
                    cur_ans = x

                    if self.cur_tool_spec is not None:
                        tool_ans = tool_invoker_fn(
                            self.cur_tool_spec["name"],
                            return_results_only=True,
                            **self.cur_tool_spec["input"],
                        )

                        # append assistant responses
                        assistant_msg = {"role": "assistant", "content": []}
                        if cur_ans is not None and cur_ans.strip() != "":
                            assistant_msg["content"].append(
                                {
                                    "type": "text",
                                    "text": cur_ans,
                                },
                            )
                        assistant_msg["content"].append(self.cur_tool_spec)

                        next_user_msg = {
                            "role": "user",
                            "content": [
                                {
                                    "type": "tool_result",
                                    "tool_use_id": self.cur_tool_spec["id"],
                                    "content": tool_ans,
                                }
                            ],
                        }
                        body["messages"].append(assistant_msg)
                        body["messages"].append(next_user_msg)

                        # keep a log of messages that had to be appended due to tool use
                        self.tool_use_added_msgs.append(assistant_msg)
                        self.tool_use_added_msgs.append(next_user_msg)
                        llm_body_changed = True

                    # TODO: Include proper token count and pricing
                    ans_word_count = len(
                        re.findall(r"\w+", postpend + cur_ans + str(self.stop_reason))
                    )
                    self.word_counts.append(
                        {
                            "request_word_count": word_count,
                            "answer_word_count": ans_word_count,
                            "price_estimate": 0.001
                            * (0.003 * word_count + 0.015 * ans_word_count),
                            "exec_time_in_s": time.time() - t0,
                        }
                    )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."

    def _response_gen(self, response_body, postpend=""):
        cur_ans = ""
        cur_tool_spec = None
        for x in response_body:
            out_dict = json.loads(x["chunk"]["bytes"])
            if "type" in out_dict.keys() and out_dict["type"] == "message_start":
                print(x)

            txt = ""
            if "content_block" in out_dict.keys():
                if out_dict["content_block"]["type"] == "text":
                    txt = out_dict["content_block"]["text"]
                elif out_dict["content_block"]["type"] == "tool_use":
                    cur_tool_spec = out_dict["content_block"].copy()
                    cur_tool_spec["input"] = ""

            elif "delta" in out_dict.keys():
                txt = out_dict["delta"].get("text", "")
                if cur_tool_spec is not None:
                    cur_tool_spec["input"] += out_dict["delta"].get("partial_json", "")

            if txt != "":
                cur_ans += txt
                yield postpend + cur_ans
            stop_reason = (
                out_dict["delta"].get("stop_reason", None)
                if "delta" in out_dict.keys()
                else None
            )
            if stop_reason is not None and stop_reason == "stop_sequence":
                stop_txt = out_dict["delta"]["stop_sequence"]
                yield postpend + cur_ans + stop_txt
                break
        if cur_tool_spec is not None:
            cur_tool_spec["input"] = json.loads(cur_tool_spec["input"])
        self.cur_tool_spec = cur_tool_spec
        self.stop_reason = stop_reason


class LLM_Mistral_Bedrock(LLM_Service):
    # https://www.promptingguide.ai/models/mixtral
    def __init__(self, bedrock_client, model_size):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """

        self.bedrock_client = bedrock_client
        if model_size == "Mixtral 8x7B v0:1":
            self.model_id = "mistral.mixtral-8x7b-instruct-v0:1"
            self.llm_description = "Mistral Mixtral 8x7B LLM from Bedrock"
            self.price_per_M_input_tokens = 0.45
            self.price_per_M_output_tokens = 0.7
        elif model_size == "Mistral Large v1":
            self.model_id = "mistral.mistral-large-2402-v1:0"
            self.llm_description = "Mistral Large LLM from Bedrock"
            self.price_per_M_input_tokens = 4
            self.price_per_M_output_tokens = 12

        self.config = {
            # "prompt": prompt,
            "max_tokens": 2000,
            "temperature": 0.5,  # 0.5 is default,
            "top_k": 50,
            "top_p": 0.9,
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        return format_messages_for_mistral(msg_list)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        tools=None,
        tool_invoker_fn=None,
        cur_fail_sleep=6,
    ):
        """
        Invokes the Llama2 large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId=self.model_id, body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking {self.llm_description}. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["outputs"][0]["text"]
                    cur_ans += partial

                    # we include those again so there's no need to leave these in the answer
                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")

                    # we have to manually check for stopword generation
                    extra_stop_break = False
                    for x in extra_stop_sequences:
                        cur_split = cur_ans.split(x)
                        if len(cur_split) > 1:
                            cur_ans = cur_split[0] + x
                            extra_stop_break = True
                            break

                    yield postpend + cur_ans

                    if extra_stop_break:
                        break
                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']

                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.00195 * word_count + 0.00256 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


class LLM_Claude2_1_Bedrock(LLM_Service):
    def __init__(self, bedrock_client):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """

        self.bedrock_client = bedrock_client
        self.llm_description = "Anthropic Claude 2.1 LLM"
        self.price_per_M_input_tokens = 8
        self.price_per_M_output_tokens = 24

        self.config = {
            # "prompt": prompt,
            "max_tokens_to_sample": 2500,
            "temperature": 0.3,  # 0.5 is default,
            "top_k": 250,
            "top_p": 1,
            "stop_sequences": ["\n\nHuman:"],
            "anthropic_version": "bedrock-2023-05-31",
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        return format_messages_for_claude(msg_list)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        cur_fail_sleep=6,
    ):
        """
        Invokes the Claude large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId="anthropic.claude-v2:1", body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking Claude. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["completion"]
                    cur_ans += partial
                    yield postpend + cur_ans
                    stop_reason = json.loads(x["chunk"]["bytes"])["stop"]
                    if (
                        stop_reason is not None
                        and stop_reason != body["stop_sequences"][0]
                    ):
                        yield postpend + cur_ans + stop_reason

                ans_word_count = len(
                    re.findall(r"\w+", postpend + cur_ans + stop_reason)
                )
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.008 * word_count + 0.024 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


class LLM_Claude_Instant_1_2_Bedrock(LLM_Service):
    def __init__(self, bedrock_client):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """

        self.bedrock_client = bedrock_client
        self.llm_description = "Anthropic Claude Instant 1.0 LLM"
        self.price_per_M_input_tokens = 0.8
        self.price_per_M_output_tokens = 2.4
        self.config = {
            # "prompt": prompt,
            "max_tokens_to_sample": 600,
            "temperature": 0.3,
            "top_k": 250,
            "top_p": 1,
            "stop_sequences": ["\n\nHuman:"],
            "anthropic_version": "bedrock-2023-05-31",
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        return format_messages_for_claude(msg_list)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        cur_fail_sleep=1,
    ):
        """
        Invokes the Claude large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html
        # start time
        t0 = time.time()

        body = self.config.copy()
        body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId="anthropic.claude-instant-v1", body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking Claude Instant. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["completion"]
                    cur_ans += partial
                    yield postpend + cur_ans
                    stop_reason = json.loads(x["chunk"]["bytes"])["stop"]
                    if (
                        stop_reason is not None
                        and stop_reason != body["stop_sequences"][0]
                    ):
                        yield postpend + cur_ans + stop_reason

                ans_word_count = len(
                    re.findall(r"\w+", postpend + cur_ans + stop_reason)
                )
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.0008 * word_count + 0.0024 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


class LLM_Llama13b(LLM_Service):
    """https://huggingface.co/blog/llama2#how-to-prompt-llama-2"""

    def __init__(self, bedrock_client):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """

        self.bedrock_client = bedrock_client
        self.llm_description = "Llama2 13b v1 LLM"
        self.price_per_M_input_tokens = 0.75
        self.price_per_M_output_tokens = 1

        self.config = {
            "max_gen_len": 1024,
            "top_p": 0.9,
            "temperature": 0.6,
        }

        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        return format_messages_for_llama(msg_list)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        cur_fail_sleep=1,
    ):
        """
        Invokes the Llama2 large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId="meta.llama2-13b-chat-v1", body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking Llama 2 chat 13b. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["generation"]
                    cur_ans += partial

                    # we include those again so there's no need to leave these in the answer
                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")

                    # we have to manually check for stopword generation
                    extra_stop_break = False
                    for x in extra_stop_sequences:
                        cur_split = cur_ans.split(x)
                        if len(cur_split) > 1:
                            cur_ans = cur_split[0] + x
                            extra_stop_break = True
                            break

                    yield postpend + cur_ans

                    if extra_stop_break:
                        break
                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']

                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.00075 * word_count + 0.001 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


class LLM_Llama70b(LLM_Service):
    """https://huggingface.co/blog/llama2#how-to-prompt-llama-2"""

    def __init__(self, bedrock_client):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        self.bedrock_client = bedrock_client
        self.llm_description = "Llama2 70b v1 LLM"
        self.price_per_M_input_tokens = 1.95
        self.price_per_M_output_tokens = 2.56

        self.config = {
            "max_gen_len": 1024,
            "top_p": 0.9,
            "temperature": 0.6,
        }

        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        return format_messages_for_llama(msg_list)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        cur_fail_sleep=1,
    ):
        """
        Invokes the Llama2 large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId="meta.llama2-70b-chat-v1", body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking Llama 2 chat 70b. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["generation"]
                    cur_ans += partial

                    # we include those again so there's no need to leave these in the answer
                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")

                    # we have to manually check for stopword generation
                    extra_stop_break = False
                    for x in extra_stop_sequences:
                        cur_split = cur_ans.split(x)
                        if len(cur_split) > 1:
                            cur_ans = cur_split[0] + x
                            extra_stop_break = True
                            break

                    yield postpend + cur_ans

                    if extra_stop_break:
                        break
                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']

                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.00195 * word_count + 0.00256 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


class LLM_Llama3(LLM_Service):
    def __init__(self, bedrock_client, model):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        self.bedrock_client = bedrock_client
        if model == "Llama3 8B Instruct - Bedrock":
            self.model_id = "meta.llama3-8b-instruct-v1:0"
            self.llm_description = "Llama3 8B Instruct LLM from Bedrock"
            self.price_per_M_input_tokens = 0.3
            self.price_per_M_output_tokens = 0.6
        elif model == "Llama3 70B Instruct - Bedrock":
            self.model_id = "meta.llama3-70b-instruct-v1:0"
            self.llm_description = "Llama3 70B Instruct LLM from Bedrock"
            self.price_per_M_input_tokens = 2.65
            self.price_per_M_output_tokens = 3.5
        elif model == "Llama3_1 70B Instruct - Bedrock":
            self.model_id = "meta.llama3-1-70b-instruct-v1:0"
            self.llm_description = "Llama3.1 70B Instruct LLM from Bedrock"
            self.price_per_M_input_tokens = 2.65
            self.price_per_M_output_tokens = 3.5
        elif model == "Llama3_1 8B Instruct - Bedrock":
            self.model_id = "meta.llama3-1-8b-instruct-v1:0"
            self.llm_description = "Llama3.1 8B Instruct LLM from Bedrock"
            self.price_per_M_input_tokens = 0.3
            self.price_per_M_output_tokens = 0.6
        elif model == "Llama3_1 405B Instruct - Bedrock":
            self.model_id = "meta.llama3-1-405b-instruct-v1:0"
            self.llm_description = "Llama3.1 405B Instruct LLM from Bedrock"
            self.price_per_M_input_tokens = 5.32
            self.price_per_M_output_tokens = 16.0

        self.config = {
            "max_gen_len": 1024,
            "top_p": 0.9,
            "temperature": 0.6,
        }

        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, messages):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        """Format messages for Llama-3 chat models.

        The model only supports 'system', 'user' and 'assistant' roles, starting with 'system', then 'user' and
        alternating (u/a/u/a/u...). The last message must be from 'user'.
        """
        prompt = ["<|begin_of_text|>"]
        assert messages[-1]["role"] == "user", "Last message has to be from user"

        for msg in messages:
            prompt.extend(
                [
                    "<|start_header_id|>",
                    msg["role"].strip(),
                    "<|end_header_id|>",
                    msg["content"].strip(),
                    "<|eot_id|>",
                ]
            )

        prompt.append("<|start_header_id|>assistant<|end_header_id|>")
        return "".join(prompt)

    def invoke_streaming(
        self,
        prompt,
        postpend="",
        extra_stop_sequences=[],
        max_retries=25,
        tools=None,
        tool_invoker_fn=None,
        cur_fail_sleep=1,
    ):
        """
        Invokes the Llama3 large-language model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        :return: Inference response from the model.
        """
        assert (
            tools is None and tool_invoker_fn is None
        ), "Native function calling not supported for Llama3"

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
        body["prompt"] = prompt + postpend

        for k in range(max_retries):
            try:
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId=self.model_id, body=json.dumps(body)
                )
                word_count = len(re.findall(r"\w+", body["prompt"]))
                print(f"Invoking {self.llm_description}. Word count: {word_count}")

                cur_ans = ""
                for x in response["body"]:
                    partial = json.loads(x["chunk"]["bytes"])["generation"]
                    cur_ans += partial

                    # we have to manually check for stopword generation
                    extra_stop_break = False
                    for x in extra_stop_sequences:
                        cur_split = cur_ans.split(x)
                        if len(cur_split) > 1:
                            cur_ans = cur_split[0] + x
                            extra_stop_break = True
                            break

                    yield postpend + cur_ans

                    if extra_stop_break:
                        break
                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']

                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.00195 * word_count + 0.00256 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
                return
            except Exception as e:
                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2
        yield "Could not invoke the AI model."


def format_messages_for_claude(messages: List[Dict[str, str]]) -> List[str]:
    """Format messages for Claude 2.1 chat models."""
    prompt: List[str] = []

    if messages[0]["role"] == "system":
        prompt.append(messages[0]["content"])
        messages = messages[1:]

    assert messages[-1]["role"] == "user", "Last message has to be from user"

    last_msg_sender = "assistant"
    for msg in messages:
        assert (
            msg["role"] != last_msg_sender
        ), "Multiple messages from the same sender in sequence"
        assert msg["role"] in (
            "user",
            "assistant",
        ), "Role has to be user, assistant or system (first message only)"
        last_msg_sender = msg["role"]
        if msg["role"] == "assistant":
            prompt.append("\n\nAssistant: " + msg["content"])
        elif msg["role"] == "user":
            prompt.append("\n\nHuman: " + msg["content"])

    prompt.append("\n\nAssistant: ")
    return "".join(prompt)


def format_messages_for_llama(messages: List[Dict[str, str]]) -> List[str]:
    """Format messages for Llama-2 chat models.

    The model only supports 'system', 'user' and 'assistant' roles, starting with 'system', then 'user' and
    alternating (u/a/u/a/u...). The last message must be from 'user'.
    """
    prompt: List[str] = []
    assert messages[-1]["role"] == "user", "Last message has to be from user"

    if messages[0]["role"] == "system":
        content = "".join(
            [
                "<<SYS>>\n",
                messages[0]["content"],
                "\n<</SYS>>\n\n",
                messages[1]["content"],
            ]
        )
        messages = [{"role": messages[1]["role"], "content": content}] + messages[2:]

    for user, answer in zip(messages[::2], messages[1::2]):
        prompt.extend(
            [
                "<s>",
                "[INST] ",
                (user["content"]).strip(),
                " [/INST] ",
                (answer["content"]).strip(),
                "</s>",
            ]
        )

    prompt.extend(["<s>", "[INST] ", (messages[-1]["content"]).strip(), " [/INST] "])

    return "".join(prompt)


def format_messages_for_mistral(messages: List[Dict[str, str]]) -> List[str]:
    """Format messages for mistral chat models.

    The model only supports 'system', 'user' and 'assistant' roles, starting with 'system', then 'user' and
    alternating (u/a/u/a/u...). The last message must be from 'user'.
    """
    prompt: List[str] = []
    assert messages[-1]["role"] == "user", "Last message has to be from user"

    if messages[0]["role"] == "system":
        content = "".join([messages[0]["content"], "\n\n", messages[1]["content"]])
        messages = [{"role": messages[1]["role"], "content": content}] + messages[2:]

    for user, answer in zip(messages[::2], messages[1::2]):
        prompt.extend(
            [
                "<s>",
                "[INST] ",
                (user["content"]).strip(),
                " [/INST] ",
                (answer["content"]).strip(),
                "</s>",
            ]
        )

    prompt.extend(["<s>", "[INST] ", (messages[-1]["content"]).strip(), " [/INST] "])

    return "".join(prompt)



import os
import numpy as np


ans = None


class ToolSolvePythonCode:
    def __init__(self):
        self.name = "solve_with_python"

        self.tool_description = {
            "name": self.name,
            "description": f"""Runs Python code to answer questions that can be easily solved with basic Python code.
Ensure that the code generates the final answer to the problem, without requiring any further analysis. Assign the final answer to the variable "ans".

Use {self.name} in any cases where a simple Python program can give the correct answer. Some examples are provided below in <use_cases></use_cases>:
<use_cases>
<use_case>String manipulation (e.g. reversing or sorting strings)<use_case>
<use_case>Sorting, ordering, filtering<use_case>
</use_cases>

Raises ValueError: if the code to be executed was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "plot_code": {
                        "type": "string",
                        "description": "Python code that, when executed, will provide the final answer to the problem",
                    },
                },
                "required": ["python_code"],
            },
        }

    def __call__(self, python_code, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        try:
            global ans
            if ans is not None:
                ans = None
            exec(python_code, globals())
        except Exception as e:
            return f"Code did NOT execute correctly.\nError description: {str(e)}"

        return str(ans)
